﻿// simple experiment having CSS animations apply to SVG
// target HTML file expects SVG line and animation rules defined as
var propertyMapping = new Array();      // holds how the SVG attributes map to temporary HTML properties
var elementWatchList = new Array();     // Identifies which elements to inspect for values during the animation
var KEYFRAMES_RULE = 7;
var KEYFRAME_RULE = 8;
var cssDOM = new Array();


function initModel() {

    try {
        createPropertyMapping();    // intialize the list of mappings
        analyzeStyleSheets();       // parse style sheets and replace SVG attributes with parallel HTML properties
        updatePropertyValues();     // start the updateLoop
    }
    catch (ex) {
        //alert(ex);
    }
}


function createPropertyMapping() {

    setAttr(propertyMapping, { SVGProperty: "x", CSSProperty: "left", CSSPropertyName: "left", units: "" });
    setAttr(propertyMapping, { SVGProperty: "y", CSSProperty: "top", CSSPropertyName: "top", units: "" });
    setAttr(propertyMapping, { SVGProperty: "cx", CSSProperty: "left", CSSPropertyName: "left", units: "" });
    setAttr(propertyMapping, { SVGProperty: "cy", CSSProperty: "top", CSSPropertyName: "top", units: "" });
    setAttr(propertyMapping, { SVGProperty: "r", CSSProperty: "minWidth", CSSPropertyName: "minWidth", units: "" });
    setAttr(propertyMapping, { SVGProperty: "width", CSSProperty: "width", CSSPropertyName: "width", units: "" });
    setAttr(propertyMapping, { SVGProperty: "height", CSSProperty: "height", CSSPropertyName: "height", units: "" });
    setAttr(propertyMapping, { SVGProperty: "fill", CSSProperty: "color", CSSPropertyName: "color", units: "" });
    setAttr(propertyMapping, { SVGProperty: "stroke-width", CSSProperty: "maxWidth", CSSPropertyName: "maxWidth", units: "" });
    setAttr(propertyMapping, { SVGProperty: "rx", CSSProperty: "width", CSSPropertyName: "width", units: "" });
    setAttr(propertyMapping, { SVGProperty: "ry", CSSProperty: "height", CSSPropertyName: "height", units: "" });
    setAttr(propertyMapping, { SVGProperty: "fill-opacity", CSSProperty: "opacity", CSSPropertyName: "opacity", units: "" });
    setAttr(propertyMapping, { SVGProperty: "startOffset", CSSProperty: "left", CSSPropertyName: "left", units: "" });
    setAttr(propertyMapping, { SVGProperty: "stroke", CSSProperty: "color", CSSPropertyName: "color", units: "" });
    setAttr(propertyMapping, { SVGProperty: "stroke-opacity", CSSProperty: "opacity", CSSPropertyName: "opacity", units: "" });
    setAttr(propertyMapping, { SVGProperty: "x1", CSSProperty: "left", CSSPropertyName: "left", units: "" });
    setAttr(propertyMapping, { SVGProperty: "x2", CSSProperty: "right", CSSPropertyName: "right", units: "" });
    setAttr(propertyMapping, { SVGProperty: "y1", CSSProperty: "top", CSSPropertyName: "top", units: "" });
    setAttr(propertyMapping, { SVGProperty: "y2", CSSProperty: "bottom", CSSPropertyName: "bottom", units: "" });
    setAttr(propertyMapping, { SVGProperty: "stop-color", CSSProperty: "color", CSSPropertyName: "color", units: "" });
    setAttr(propertyMapping, { SVGProperty: "stop-opacity", CSSProperty: "opacity", CSSPropertyName: "opacity", units: "" });
    setAttr(propertyMapping, { SVGProperty: "lighting-color", CSSProperty: "color", CSSPropertyName: "color", units: "" });
    setAttr(propertyMapping, { SVGProperty: "offset", CSSProperty: "opacity", CSSPropertyName: "opacity", units: "" });
    setAttr(propertyMapping, { SVGProperty: "flood-color", CSSProperty: "color", CSSPropertyName: "color", units: "" });
    setAttr(propertyMapping, { SVGProperty: "flood-opacity", CSSProperty: "opacity", CSSPropertyName: "opacity", units: "" });
    setAttr(propertyMapping, { SVGProperty: "z", CSSProperty: "zIndex", CSSPropertyName: "zIndex", units: "" });
    setAttr(propertyMapping, { SVGProperty: "bias", CSSProperty: "opacity", CSSPropertyName: "opacity", units: "" });

    setAttr(propertyMapping, { SVGProperty: "dx", CSSProperty: "marginLeft", CSSPropertyName: "marginLeft", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "dy", CSSProperty: "marginTop", CSSPropertyName: "marginTop", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "specularConstant", CSSProperty: "width", CSSPropertyName: "width", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "specularExponent", CSSProperty: "height", CSSPropertyName: "height", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "diffuseConstant", CSSProperty: "width", CSSPropertyName: "width", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "surfaceScale", CSSProperty: "height", CSSPropertyName: "height", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "limitingConeAngle", CSSProperty: "maxWidth", CSSPropertyName: "maxWidth", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "pointsAtX", CSSProperty: "paddingLeft", CSSPropertyName: "paddingLeft", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "pointsAtY", CSSProperty: "paddingTop", CSSPropertyName: "paddingTop", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "pointsAtZ", CSSProperty: "paddingBottom", CSSPropertyName: "paddingBottom", units: "" });
    setAttr(propertyMapping, { SVGProperty: "k1", CSSProperty: "minWidth", CSSPropertyName: "minWidth", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "k2", CSSProperty: "maxWidth", CSSPropertyName: "maxWidth", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "k3", CSSProperty: "minHeight", CSSPropertyName: "minHeight", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "k4", CSSProperty: "maxHeight", CSSPropertyName: "maxHeight", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "fx", CSSProperty: "marginRight", CSSPropertyName: "marginRight", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "fy", CSSProperty: "marginBottom", CSSPropertyName: "marginBottom", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "slope", CSSProperty: "height", CSSPropertyName: "height", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "intercept", CSSProperty: "width", CSSPropertyName: "width", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "exponent", CSSProperty: "left", CSSPropertyName: "left", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "amplitude", CSSProperty: "right", CSSPropertyName: "right", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "azimuth", CSSProperty: "height", CSSPropertyName: "height", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "elevation", CSSProperty: "width", CSSPropertyName: "width", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "scale", CSSProperty: "lineHeight", CSSPropertyName: "lineHeight", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "stroke-dashoffset", CSSProperty: "left", CSSPropertyName: "left", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "targetx", CSSProperty: "left", CSSPropertyName: "left", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "targety", CSSProperty: "top", CSSPropertyName: "top", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "divisor", CSSProperty: "minWidth", CSSPropertyName: "minWidth", units: "px" });
    setAttr(propertyMapping, { SVGProperty: "numOctaves", CSSProperty: "lineHeight", CSSPropertyName: "lineHeight", units: "px" });


}


// store property mappings
function setAttr(attrArray, properties) {

    // may expand this function to do more than store properties;  if I don't I will just
    // push the values in createPropertyMapping
    attrArray.push(properties);
}


function analyzeStyleSheets() {

    // walk through each stylesheet
    var styleString = "";
    var currentCSSStyleSheet;
    var currentCSSRule;
    var curentCSSStyle;
    for (var currentSheet = 0; currentSheet < document.styleSheets.length; currentSheet++) {
        currentCSSStyleSheet = document.styleSheets.item(currentSheet);
        // walk through each rule

        jQuery.get(currentCSSStyleSheet.href.toString(), null, function (data) {
            parseCSS(data);

            for (var currentRule = 0; currentRule < currentCSSStyleSheet.cssRules.length; currentRule++) {
                currentCSSRule = currentCSSStyleSheet.cssRules[currentRule];
                if (currentCSSRule.type == KEYFRAME_RULE)
                // keyframe rule
                {
                    setKeyFrameRules(currentCSSRule);
                }
                else if (currentCSSRule.style != null) {
                    // if there is at least one valid SVG rule that needs to be converted, do so and create a <div> on which to store them
                    //createStructureFromRule(currentCSSRule, currentCSSRule.style.cssText);
                    createStructureFromRule(currentCSSRule, getAllCssRuleText(currentCSSRule.selectorText, null));
                }
            }
        });
    }
}


function setKeyFrameRules(cssRule) {

    // change the rule to the corresponding css property, e.g. x -> left
    for (var i = 0; i < cssRule.cssRules.length; i++) {
        //var arrayStyle = parseRuleText(cssRule, cssRule.cssRules[i].style.cssText);
        var arrayStyle = parseRuleText(cssRule, getAllCssRuleText(cssRule.name, cssRule.cssRules[i].keyText));
        for (var j = 0; j < arrayStyle.length; j++) {
            var ind = findPropertyMappingIndex(arrayStyle[j][0], "SVG");
            if (ind != -1) {
                var CSSProperty = propertyMapping[ind].CSSPropertyName;
                cssRule.cssRules[i].style[CSSProperty] = propertyMapping[ind].units ? parseFloat(arrayStyle[j][1]) + propertyMapping[ind].units : arrayStyle[j][1];
            }
        }
    }
}


// create a DIV with the styles converted from the SVG Element (i.e. X -> Left)
function createStructureFromRule(cssRule, cssRuleText) {

    if (cssRule.type == 1) {
        // fragile and too simple, presume this is for SVG and create a div of the same class
        var className = cssRule.selectorText.substr(1);
        var div = createDiv(className);
        var styleArray = parseRuleText(cssRule, cssRuleText);
        elementWatchList.push(new elementWatch(div, className, styleArray));
    }
}


// consider making this a dynamic inline function
function elementWatch(element, className, styleArray) {

    this.element = element;
    this.className = className;
    this.styleArray = styleArray;
    return this;
}


// split CSSRuleText into an array of keypairs of property and value;
function parseRuleText(cssRule, cssRuleText) {

    var arrayPropertySet = cssRuleText.toString().trim().split(";");
    for (var i = 0; i < arrayPropertySet.length; i++) {

        arrayPropertySet[i] = arrayPropertySet[i].toString().trim().split(":");
        try {
            var ind = findPropertyMappingIndex(arrayPropertySet[i][0], "SVG");
            if (ind != -1) {
                var a = propertyMapping[ind].CSSProperty;
                cssRule.style[a] = propertyMapping[ind].units ? parseFloat(arrayPropertySet[i][1]) + propertyMapping[ind].units : arrayPropertySet[i][1];
            }
        }
        catch (e) {
            //alert(e.message);
        }
    }

    return arrayPropertySet;
}


// do i need property set?
function createDiv(className, propertySet) {

    // do i need property set?
    var newDiv = document.createElement('div');
    // apply the class name so this Div can participate in the real animation
    newDiv.setAttribute('class', className);
    // set position abosolute for left/top stick (issue here)
    newDiv.setAttribute("style", "position:absolute;visibility:hidden");
    document.body.appendChild(newDiv);
    return newDiv;
}


function updatePropertyValues() {

    // walk through each hydrated element
    for (var i = 0; i < elementWatchList.length; i++) {

        // get the div being animated
        var div = elementWatchList[i].element;

        // match this to any other element that shares that class
        // TODO : make these SVG only
        var elementArray = document.getElementsByClassName(elementWatchList[i].element.className);

        // walk throug each element that shares this class
        for (var j = 0; j < elementArray.length; j++)

        // for each style that is being tracked
        // if it is not a div
            if (elementArray[j].nodeName.toLowerCase() != "div") {
                for (var k = 0; k < elementWatchList[i].styleArray.length; k++) {
                    //var styleToTranslate = elementWatchList[i].styleArray[k];
                    // currently not stripping out -ms- which causes a failure here
                    try {
                        var ind = findPropertyMappingIndex(elementWatchList[i].styleArray[k][0], "SVG");
                        if (ind != -1) {
                            var HTMLElement = propertyMapping[ind].CSSPropertyName;
                            var value = propertyMapping[ind].units ? parseFloat(document.defaultView.getComputedStyle(elementWatchList[i].element, null)[HTMLElement]) : (document.defaultView.getComputedStyle(elementWatchList[i].element, null)[HTMLElement]);
                            elementArray[j].setAttribute(elementWatchList[i].styleArray[k][0], value);
                            elementArray[j].style[elementWatchList[i].styleArray[k][0]] = value;
                        }
                    }
                    catch (e) {
                    }
                }
            }
    }

    // Future-proof: when feature is fully standardized
    if (window.requestAnimationFrame) window.requestAnimationFrame(updatePropertyValues);
    // IE implementation
    else if (window.msRequestAnimationFrame) window.msRequestAnimationFrame(updatePropertyValues);
    // Firefox implementation
    else if (window.mozRequestAnimationFrame) window.mozRequestAnimationFrame(updatePropertyValues);
    // Chrome implementation
    else if (window.webkitRequestAnimationFrame) window.webkitRequestAnimationFrame(updatePropertyValues);
    // Other browsers that do not yet support feature
    else
        setTimeout(updatePropertyValues, 60);
}


// use to find related properties
// NOTE : I conisdered doing a two key array here but the cost simply wasn't worth the simplicity
function findPropertyMapping(property, type, style) {

    // this could be slightly more efficient if I created a two key'd index array into the other array
    for (var i = 0; i < propertyMapping.length; i++) {
        if (type == "SVG" && propertyMapping[i].SVGProperty == property)
            if (style)
                return propertyMapping[i].CSSPropertyName;
            else
                return propertyMapping[i].CSSProperty;

        if (type == "HTML" && propertyMapping[i].CSSProperty == property)
            return propertyMapping[i].SVGProperty;
    }

    return null;
}


function findPropertyMappingIndex(property, type) {

    // this could be slightly more efficient if I created a two key'd index array into the other array
    for (var i = 0; i < propertyMapping.length; i++) {
        if (type == "SVG" && propertyMapping[i].SVGProperty == property)
            return i;

        if (type == "HTML" && propertyMapping[i].CSSProperty == property)
            return i;
    }

    return -1;
}


// workaround for webkit
function getAllCssRuleText(name, subName) {

    for (var i = 0; i < cssDOM.length; i++) {
        if (cssDOM[i].selector == name) {
            if (subName) {
                if (cssDOM[i].subSelector == subName.replace(/%$/ig, '').split('.')[0]) {
                    return cssDOM[i].properties;
                }
            }
            else {
                return cssDOM[i].properties;
            }
        }
    }

    return null;
}


// workaround for webkit
function parseCSS(cssData) {

    var splitExp = /{|}/;
    var lines = cssData.replace(/(?:^|\s)(from)(?=\s|$)/ig, '0%').replace(/(?:^|\s)(to)(?=\s|$)/ig, '100%').replace(/[\n|\r]/g, ' ').replace(/\t*/g, '').replace(/\s{2}/, ' ').split(splitExp);
    var ind = 0;
    var line;

    for (var ind = 0; ind < lines.length; ++ind) {

        var selector;
        var subSelector;
        var props;
        var line = trim(lines[ind]);

        if (!line)
            continue;

        if (isAnimSelector(line)) {
            subSelector = line.replace(/%$/ig, '');
        }
        else {
            if (isSelector(line)) {
                selector = trim(line.substring(line.indexOf(' ')));
            }
            else {
                props = line;
                cssDOM.push(new cssRuleElement(selector, subSelector, props));
                subSelector = props = '';
            }
        }
    }

    function trim(str) {
        return (str || '').replace(/^\s\s*/, '').replace(/\s\s*$/, '');
    };


    function isSelector(value) {
        return (value && (!(/;/ig.test(value))));
    };

    function isAnimRule(value) {
        return (value && (/^@-/ig.test(value)));
    };

    function isAnimSelector(value) {
        return (value && (/%$/ig.test(value)));
    };

    function cssRuleElement(selector, subSelector, properties) {
        this.selector = selector;
        this.subSelector = subSelector;
        this.properties = properties;
        return this;
    };

    return cssDOM;
}
